#include "ScannerSModelclasses.h"
#include "ScannerSAux.h"
#ifdef VERBOSE
#include "ScannerSVerbose.h"
#endif

using namespace std;

//Routine to generate a VEV point uniformly
void PhiVec::generate(){
  for(vector<double>::iterator iter=Vec.begin(),iter1=Vecmin.begin(),iter2=Vecmax.begin();iter!=Vec.end();)
    *iter++=r(*iter1++,*iter2++);//generate uniformly
}

//Routine to generate a VEV point uniformly
void PhiVec::generate(double (*pdf)(const std::vector<double> &)){
  const unsigned nmax=100000;
  unsigned i=0;
  for(;i!=nmax;++i){
    for(vector<double>::iterator iter=Vec.begin(),iter1=Vecmin.begin(),iter2=Vecmax.begin();iter!=Vec.end();)
      *iter++=r(*iter1++,*iter2++);//generate uniformly
    if(r()<pdf(Vec))
      break;
  }
  if(i>=nmax){
    cerr << "ERROR in void PhiVec::generate(double (*pdf)(const std::vector<double> &)): Too many tries. nmax = "<<nmax << endl; 
    exit(-1);
  }

}

void PhiVec::setphi(const std::vector<double> & phi_in){
  if(phi_in.size()==Vec.size())
    Vec=phi_in;
  else
    std::cerr<< "ERROR in PhiVec::setphi(const std::vector<double> & phi_in) - vector sizes don't match. Set failed."<< endl;
}

std::istream & operator>>(std::istream & in, Potential & V){
  if(V.Mon.size()!=0)
    clog<< "Warning! Your matrix of monomials was not empty. Clearing matrix before proceeding..."<< endl;
  V.Mon.clear();
  in>>V.Phi>>V.PhiPar>>V.MixingPars>>V.MLpars>>V.Cp;
  for(Potential::size_type i=0;i!=V.Cp.size();++i){
    Potential::size_type Nmon;
    in>>Nmon;
    std::vector<Monomial> VecMon;
    Monomial M_in(V.Phi.size());
    for(Potential::size_type j=0;j!=Nmon; ++j){
      in>>M_in;
      VecMon.push_back(M_in);
    }
    V.Mon.push_back(VecMon);  
  }

  //temporary vectors to build matrices
  vector<double> tempV(V.Cp.size(),0);
  vector< vector<double> > tempdV(V.Phi.size(),tempV);
  vector< vector< vector<double> > > tempd2V(V.Phi.size(),tempdV);
  V.dV=tempdV;
  V.d2V=tempd2V;
  V.d2Vnew=tempd2V;
  V.d2V.Mag=tempV;
  V.d2Vnew.Mag=tempV;
  vector<unsigned int> tempOrder(V.Cp.size(),0);
  V.NIC=0;
  V.orderdV=tempOrder;
  V.orderd2V.resize(V.Cp.size()+V.Phi.size());
  V.TempM =  new double [V.Phi.size()*V.Phi.size()];
  V.TempM2 =  new double [V.Phi.size()*V.Phi.size()];
  V.TempOut = new double [(V.Phi.size()+1)*V.Phi.size()];
  vector<double> tempVphi(V.Phi.size(),0);
  vector< vector<double> > tempMatphi(V.Phi.size(),tempVphi);
  vector<unsigned int> tempBlocks(V.Phi.size(),0);
  vector<int> tempBlocks2(V.Phi.size(),0);
  V.BlockInfo=tempBlocks2;
  V.BlockPermutLabels=tempBlocks;
  V.Mbasis=tempMatphi;
  V.Maux=tempMatphi;
  V.Mrot=tempMatphi;
  V.Mrot2=tempMatphi;
  V.Mmixing=tempMatphi;
  V.a0tree.resize((V.Phi.size()*(V.Phi.size()+1))/2);
  for(size_t i=0;i!=(V.Phi.size()*(V.Phi.size()+1))/2;++i){
    V.a0tree[i].resize((V.Phi.size()*(V.Phi.size()+1))/2);
    for(size_t j=0; j!=(V.Phi.size()*(V.Phi.size()+1))/2;++j)
      V.a0tree[i][j].resize(V.Cp.size());
  }
  V.a0treeIni();
  V.a0treeSum = new double [(V.Phi.size()*(V.Phi.size()+1))/2*(V.Phi.size()*(V.Phi.size()+1))/2];

  return in;
} //Overloaded input operator to initialise

std::ostream & operator<<(std::ostream & os, const Potential & V){

  os<< " V(";
  for(Potential::size_type i=0;i!=V.Phi.size();++i){
    if(i!=0)
      os<<",";
    os<<"phi"<<i;
  }
  os<<") = ";

  for(Potential::size_type a=0; a!=V.Mon.size();++a){
    if(a!=0)
      os<<" + ";
    os<< "L"<<a<<"(";
    for(Potential::size_type b=0; b!=V.Mon[a].size();++b){
      if(b!=0)
	if(V.Mon[a][b].coeff_get()>0)
	  os<<" + ";
	else
	  os<<" ";
      os<<V.Mon[a][b];
    }
    os<<")";
  }

  return os;  
}

ostream & operator<<(std::ostream & os, const PhiRef & P){
  os << P.Phi_;
  return os;
}

ostream & operator<<(std::ostream & os, const LambdaRef & L){
  os << L.L_;
  return os;
}

double Potential::operator()(std::vector< std::vector<Monomial> >::const_iterator aiter) const{ 
  //Routine to compute the Potential term directly from the polynomial expression

  double subsum=0e0;
  for(std::vector<Monomial>::const_iterator miter = (*aiter).begin(); miter!=(*aiter).end();++miter){
    double prod=(*miter).coeff_get();
    for(size_type j=0; j!=Phi.size(); ++j)
      {
	// Multiply by the appropriate power of the field if the power is positive (otherwise we assume the field is not present in the monomial)
	if((*miter)[j]>0)
	  prod*=pow(Phi[j],(*miter)[j]);
      }
    // Add monomial to the sum
    subsum+=prod;
  }

  return subsum;
}

double Potential::operator()(size_type a) const{ 
  std::vector< std::vector<Monomial> >::const_iterator aiter = Mon.begin()+a;
  if((a+1)>0 && a<Mon.size())
    return (*this)(aiter);
  else{
    cerr<< "ERROR in function: double Potential::operator[](size_type a) const -- Out of bounds, a= "<< a << " dimension = "<<Mon.size() << endl;
    exit(-1);
  }
}

double Potential::operator()() const{ 
  //Routine to compute the Potential directly from the polynomial expression
  // Start sum of polynomial terms
  double sum=0e0;
  for(std::vector< std::vector<Monomial> >::const_iterator aiter = Mon.begin();aiter!=Mon.end();++aiter){
    double subsum=(*this)(aiter-Mon.begin());
    sum+=subsum * Cp[aiter-Mon.begin()];
  }
  return sum;
}

double Potential::operator()(size_type i, size_type a) const{
    //Routine to compute the derivative of the Potential directly from the polynomial expression for the a_th component.

  //Start sum of monomials  
  double sum=0e0;  
  //Loop over  all terms for the a_th component 
  for(size_type k=0; k!= Mon[a].size();++k){
    //Start loop to multiply all powers in monomial
    // First multiply by numerical factor...
    double prod=Mon[a][k].coeff_get();
    // Then if the power for the field with respect to which we are differentiating is positive multiply by the corresponding term
    if(Mon[a][k][i]>0){
      for(size_type j=0; j!=Mon[a][k].size(); ++j)
	{
	  //If we are looking at the field which we are differentiating w.r.t...
	  if(j==i){
	    prod*=Mon[a][k][j]*pow(Phi[j],Mon[a][k][j]-1);
	  }//Otherwise multiply by corresponding term...
	  else if(Mon[a][k][j]>0)
	    prod*=pow(Phi[j],Mon[a][k][j]);
	}
    }//Otherwise mutiply by zero because the monomial doesn't depend on the field
    else
      prod=0e0;
    //Sum monomials...
    sum+=prod;
  }
  //Return final value for derivative...
  return sum;
}

double Potential::operator()(size_type i, size_type j,size_type a) const{
  //Routine to compute the second derivative of the Potential directly from the polynomial expression for the a_th component.

  //Start sum of terms...
  double sum=0e0;
  for(size_type k=0; k!=Mon[a].size();++k){
    //Start products
    double prod=Mon[a][k].coeff_get();
    //Case of second order derivative w.r.t. same phi
    if(i==j){
      // Do as before and compute if power is large enough for phi being deffirentiated w.r.t.
      if(Mon[a][k][j]>1){
	for(size_t j1=0; j1!=Mon[a][k].size(); ++j1)
	  {
	    // Field beig differentiated...
	    if(j1==j){
	      prod*=Mon[a][k][j1]*(Mon[a][k][j1]-1)*pow(Phi[j1],Mon[a][k][j1]-2);
	    }//Other fields
	    else if(Mon[a][k][j1]>0)
		prod*=pow(Phi[j1],Mon[a][k][j1]);
	  }
      }// Set to zero if power not large enough
      else
	prod=0e0;
    }//Mixed derivatives case
    else {
      //Check that powers are large enough for both fields in the mixed derivative...
      if(Mon[a][k][j]>0 && Mon[a][k][i]>0){
	for(size_type j1=0; j1!=Mon[a][k].size(); ++j1)
	  {//Fields being differentiated case...
	    if(j1==j || j1== i){
	      prod*=Mon[a][k][j1]*pow(Phi[j1],Mon[a][k][j1]-1);
	    }//Other fields...
	    else if(Mon[a][k][j1]>0)
		prod*=pow(Phi[j1],Mon[a][k][j1]);
	  }
      }// Not enough powers...
      else
	prod=0e0; 
    }
    //Sum all terms...
    sum+=prod;
  }
  return sum;
}



bool Potential::generate(){
  int hcont=0;
  do{
    if(PhiPar.size()>0){
      PhiPar.generate();
      MyPhiParametrization(PhiPar,Phi);
    }
    else{//generate Phi
      Phi.generate();
    }
    //Compute first derivative matrix
    for(vector< vector<double> >::iterator iter=dV.begin();iter!=dV.end();++iter)
      for(vector<double>::iterator ater=iter->begin();ater!=iter->end();++ater)
	*ater=(*this)(iter-dV.begin(),ater-iter->begin());
    
#ifdef DEBUG
    /////Check: Print out matrix (in verbose mode)
    clog << endl << "**************** Print out dV ******************";
    printdV();
#endif
    
    //Perform Step 1 of gaussian elimination
    hcont=GaussUpTriang(dV);
#ifdef DEBUG
    /////Check: Print out matrix (in verbose mode)
    clog << endl << "****** Print out dV after step 1 of gaussian elimination ********";
    printdV();
#endif
  }while(hcont==0 && dV.size()>=dV[0].size() && (clog<<"WARNING: Massless theory with kinetic terms only generated. Trying again..."<< endl));
  
  //Perform step 2 of Gaussian elimination
  GaussSolveUpTriang(dV,NIC,orderdV,hcont);
#ifdef DEBUG
  /////Check: Print out matrix (in verbose mode)
  clog << endl << "****** Print out dV after step 2 of gaussian elimination ********";
  printdV();
  clog << "Number of independent linear conditions = " << NIC << endl;
  clog << endl<< "******** Ordering of couplings ***********"<< endl;
  for(size_t a=0;a!=dV[0].size();a++){
    clog <<" dV[" << a<<"] = " <<orderdV[a] << endl;
  }
#endif

  //Compute second derivative matrices
  for(vector< vector< vector<double> > >::iterator iter=d2V.begin();iter!=d2V.end();++iter)
    for(vector< vector<double> >::iterator jter=iter->begin();jter!=iter->end();++jter)
      for(vector<double>::iterator ater=jter->begin();ater!=jter->end();++ater)
	*ater=(*this)(iter-d2V.begin(),jter-iter->begin(),ater-jter->begin());

#ifdef DEBUG
    /////Check: Print out matrix (in verbose mode)
    clog << endl << "**************** Print out d2V[a][i][j] ******************";
    printd2V();
    clog << "Analysing quadratic vacuum conditions ..." << endl; 
#endif

    RedBydV(dV,d2V,NIC,orderdV);
#ifdef DEBUG
    ///////////////!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!////////////
    clog << endl << "****** d2V (Hessians) reduced by linear condition *******" << endl;
    PrintM21(d2V,NIC,orderdV);
    clog << endl<< "****** d2V print out for comparison with initial matrices ******" << endl;
    PrintM22(d2V,NIC,orderdV);
  
    clog << endl<< endl << "****** Checking for a priori eigen directions ..." << endl;
#endif

    if(FindMixing()){

#ifdef VERBOSE
      clog << endl ;
      clog << endl << "********************************************************" ;
      clog << endl << "********************************************************" ;
      clog << endl << "******** SUMMARY INFO FOR THIS ATTEMPTED POINT *********" ;
      clog << endl << "********************************************************" ;
      clog << endl << "********************************************************" << endl;
      clog << endl << "------------------";
      clog << endl << "--- New basis ----";
      clog << endl << "------------------" << endl;
    
      clog << endl << "--- Blocks which will mix ---" << endl;
      int sumsizes=0;
      for(size_t a=0;a!=BlockDims.size();++a){
	clog<< "*** Block "<<a<< endl;
	for(size_t i=sumsizes; i!=BlockDims[a]+sumsizes;++i){
	  clog << "v["<<i<<"]= ";
	  for(size_t j=0;j!=d2V.size();++j)
	    clog<<Mbasis[i][j]<< " dphi"<<j<<" + ";
	  
	  clog << endl;
	}
	sumsizes+=BlockDims[a];
      }
      int ndegen=0;
      for(size_t i=0;i!=degeneracies.size();++i)
	ndegen+=degeneracies[i];
      clog << endl << "--- Non-degenerate Curved diagonal directions ---" << endl;
      for(size_t i=Nmixed; i!=Nmixed+NnoMixCurv-ndegen; ++i){
	clog << "v["<<i<<"]= ";
	for(size_t j=0;j!=d2V.size();++j)
	  clog<<Mbasis[i][j]<< " dphi"<<j<<" + ";
	clog << endl;
      }
      clog << endl << "--- degenerate Curved diagonal directions ---" << endl;
      int sumdeg=0;
      for(size_t b=0; b!=degeneracies.size();++b){
	clog<< "*** Group "<< b<< endl;
	for(size_t i=Nmixed+NnoMixCurv-ndegen+sumdeg; i!=Nmixed+NnoMixCurv-ndegen+sumdeg+degeneracies[b]; ++i){
	  clog << "v["<<i<<"]= ";
	  for(size_t j=0;j!=d2V.size();++j)
	    clog<<Mbasis[i][j]<< " dphi"<<j<<" + ";
	  clog << endl;
	}
	sumdeg+=degeneracies[b];
      }

      clog << endl << "--- Goldstone directions ---" << endl;
      for(size_t i=Nmixed+NnoMixCurv; i!=d2V.size(); ++i){
	clog << "v["<<i<<"]= ";
	for(size_t j=0;j!=d2V.size();++j){
	  clog<<Mbasis[i][j]<< " dphi"<<j<<" + ";
	}
	clog << endl;
      }
      clog<<endl;
      clog<<"---------------------------------------------------------------"<< endl;
      clog<<"--- Independent parameters generated after VEVs and Mixings ---"<< endl;
      clog<<"---------------------------------------------------------------"<< endl<< endl;
#endif

      unsigned int NL=d2V[0][0].size()-NIC+Nmixed+NnoMixCurv;
      vector<double> Ltemp(NL);
      Cp.Lindp.resize(0);
      Cp.Mindp.resize(0);
      ///////////////////////////////////
      for(size_t A=NIC2; A!=NL;++A){
	if(orderd2V[A]<d2V[0][0].size()-NIC){
	  size_t ilambda=orderdV[NIC+orderd2V[A]];
	  Cp.L[ilambda]=r(Cp.L.vgetmin(ilambda),Cp.L.vgetmax(ilambda));
	  Ltemp[orderd2V[A]]=Cp.L[ilambda];
	  Cp.Lindp.push_back(ilambda);
	}
	else{
	  size_t iphi=orderd2V[A]-(d2V[0][0].size()-NIC);
	  Cp.Mass[iphi]=r(Cp.Mass.vgetmin(iphi),Cp.Mass.vgetmax(iphi));
	  Ltemp[orderd2V[A]]=pow(Cp.Mass[iphi],2);
	  Cp.Mindp.push_back(iphi);
	}
      }
#ifdef VERBOSE
      for(size_t A=NIC2; A!=NL;++A){
	if(orderd2V[A]<d2V[0][0].size()-NIC){
	  size_t ilambda=orderdV[NIC+orderd2V[A]];
	  clog<<"Coupling "<<ilambda<< " is independent."<< endl;
	}
	else{
	  size_t iphi=orderd2V[A]-(d2V[0][0].size()-NIC);
	  clog<<"Mass of state "<<iphi<< " is independent." <<endl;
	}
      }
#endif

      //Check if reparametrization of independent Masses/Couplings is active
      if(MLpars.size()>0){
	vector<double> IndependentMs, IndependentLs,Lvec(d2V[0][0].size()),Mvec(d2V.size());//Auxiliary variables
	//Compute re-parametrized fields
	for(size_t i=0; i!=MLpars.size();++i){
	  MLpars[i]=r(MLpars.vgetmin(i),MLpars.vgetmax(i));//Generate parameters uniformly in range decided in mathematica file
	}
	//Compute new Masses and Couplings with the user defined function 

	MyCoupMassRelations(PhiPar,Phi,Mmixing,MixingPars,Lvec,Mvec,MLpars,IndependentMs,IndependentLs,r);
	bool testML=true; //Variable to test if the choice of masses and couplings to re-parametrize is ok. Start by assuming it is and flip tp false whenever a choice is not possible
	//Check if choice is ok for masses
	for(size_t i=0;i!=IndependentMs.size();++i){
	  bool testM=false;//Assume the test will fail
	  for(size_t j=0;j!=Cp.Mindp.size();++j){
	    if(IndependentMs[i]==Cp.Mindp[j])
	      testM=true;//If the number matches one of the independent masses flip to true
	  }
	  testML*=testM; //if no match is found testM will be false so testML will also become false
	}
	//Check if choice is ok for Couplings (same procedure!)
	for(size_t i=0;i!=IndependentLs.size();++i){
	  bool testL=false;
	  for(size_t j=0;j!=Cp.Lindp.size();++j){
	    if(IndependentMs[i]==Cp.Mindp[j])
	      testL=true;
	  }
	  testML*=testL;
	}
	if(testML){//If test succeeds, then asign new re-parametrized mass & coupling values to the corresponding independent coupplings 
	  for(size_t i=0; i!=IndependentMs.size();++i)
	    Cp.Mass[IndependentMs[i]]=Mvec[IndependentMs[i]];
	  for(size_t i=0; i!=IndependentLs.size();++i)
	    Cp.L[IndependentLs[i]]=Lvec[IndependentLs[i]];
	}
      }

      //Set Goldstone masses to zero
      for(size_t ip=Nmixed+NnoMixCurv;ip!=d2V.size();++ip)
	Cp.Mass[ip]=0;

      // Determine dependent couplings
      for(size_t A=0;A!=NIC2; ++A){
	if(orderd2V[A]<d2V[0][0].size()-NIC){
	  size_t ilambda=orderdV[NIC+orderd2V[A]];
	  double sum=0e0;
	  for(size_t j=NIC2;j!=NL;++j)
	    sum-=d2Vreduced[A][orderd2V[j]]*Ltemp[orderd2V[j]];
	  
	  Cp.L[ilambda]=sum;
	}
	else{
	  size_t iphi=orderd2V[A]-(d2V[0][0].size()-NIC);
	  double sum=0e0;
	  for(size_t j=NIC2;j!=NL;++j){
	    sum-=d2Vreduced[A][orderd2V[j]]*Ltemp[orderd2V[j]];
	  }
	  if(sum<0)
	    return false;
	  Cp.Mass[iphi]=sqrt(abs(sum));
	}
      } 
      
      for(size_t a=0;a!=NIC;++a){
	size_t ilambda = orderdV[a];
	double sum=0e0;
	for(size_t j=NIC;j!=d2V[0][0].size();++j){
	  sum-=dV[a][orderdV[j]]*Cp.L[orderdV[j]];
	}
	Cp.L[ilambda]=sum;    
      }

      
      //Compute mixing matrix
      ////Part 1 which involves Mrot block
      for(size_t i=0;i!=Nmixed; ++i){
	for(size_t j=0;j!=d2V.size(); ++j){
	  Mmixing[i][j]=0e0;
	  for(size_t k=0;k!=Nmixed;++k){
	    Mmixing[i][j]+=Mrot[i][k]*Mbasis[k][j];
	  }
	}
      }
      ////Part 2 where Mrot is the identity
      for(size_t i=Nmixed;i!=d2V.size(); ++i){
	for(size_t j=0;j!=d2V.size(); ++j){
	  Mmixing[i][j]=Mbasis[i][j];
	}
      }

#ifdef VERBOSE
      clog << endl << "----------------------";
      clog << endl << "--- Mixing matrix ----";
      clog << endl << "----------------------";
      clog << endl << "Note: M_{jk} -> indices j (lines) are mass eigenstates and k (columns) are original states" << endl;

      PrintM1(Mmixing);
      clog << endl << "----------------------";
      clog << endl << "------ Masses --------";
      clog << endl << "----------------------";
      clog << endl << "Note: Mass eigenstates ordered as in the description above (see NEW BASIS section above for block basis of states that were mixed, or states that were found to be eigenstates a priori)" << endl;
      for(int i=0; i<Cp.Mass.size();++i)
	clog<< "M["<<i<<"]="<<Cp.Mass[i]<< endl;

      clog << endl << "----------------------";
      clog << endl << "------ Couplings -----";
      clog << endl << "----------------------" << endl;

      for(int i=0; i<Cp.L.size();++i)
	clog<< "L["<<i<<"]="<<Cp.L[i]<< endl;

      clog << endl << "----------------------";
      clog << endl << "-------- VEVs --------";
      clog << endl << "----------------------" << endl;

      for(int i=0; i<Phi.size();++i)
	clog<< "Phi["<<i<<"]="<<Phi[i]<< endl;


#endif
      return true;
    }
    else
      return false;

}

bool Potential::FindMixing(){
  unsigned int kflat=0; //Counts number of flat directions
  unsigned int kcurved=0; //Counts number of flat directions

  FindAprioriEigenDirs(d2V,Mbasis,TempM,TempM2,TempOut,orderdV,degeneracies,BlockInfo,BlockDims,kflat,kcurved,NIC,r);

#ifdef DEBUG
    //Print out////////!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!////////////
  clog << endl << "Found " << kflat<< " a priori flat directions" ;
  for(size_t k=d2V.size()-kflat;k!=d2V.size(); ++k){
    //Print out flat direction
    clog << endl;
    for(int i=0;i!=d2V.size();i++){
      clog << Mbasis[k][i]<< "\t"; 
    }
    if(k<d2V.size()-1){
      double dot=0;
      clog<< endl;
      for(int i=0;i!=d2V.size();i++){
	dot+=Mbasis[k][i]*Mbasis[k+1][i]; 
      }
      clog<<"dot product with next vector is "<< dot<< endl;
    }
  }
  clog << endl;
  clog << endl << "Found " << kcurved << " a priori curved eigen-directions" << endl ;
  
  for(int k=d2V.size()-kflat-kcurved;k!=d2V.size()-kflat; ++k){
    clog << endl;
    for(int i=0;i!=d2V.size();++i)
      clog << Mbasis[k][i]<< "\t"; 
  }
  clog << endl;
  clog<< "degeneracy info vector:"<< endl;
  for(size_t i=0; i<degeneracies.size();++i)
    clog<<degeneracies[degeneracies.size()-i-1]<< "\t";

  clog << endl << "Found " << d2V.size()-kflat-kcurved << " mixing directions" << endl ;
  
  for(int k=0;k!=d2V.size()-kflat-kcurved; ++k){
    clog << endl;
    for(int i=0;i!=d2V.size();++i)
      clog << Mbasis[k][i]<< "\t"; 

    if(k<d2V.size()-kflat-kcurved-1){
      double dot=0;
      clog<< endl;
      for(int i=0;i!=d2V.size();i++){
	dot+=Mbasis[k][i]*Mbasis[k+1][i]; 
      }
    }

  }
  clog << endl;

#endif
  ConvertEigen(d2V,d2Vnew,Mbasis,orderdV,kflat,kcurved,NIC);

  
#ifdef DEBUG
    ////////////////////////CHECK!!!!!!!!!!!!!!!!!!!!
  clog << endl<< "****** d2Vnew independent matrices in new basis  ******" << endl;
  PrintM22(d2Vnew,NIC,orderdV);
#endif
  //exit(-1);
  unsigned int kmix=d2V.size()-kflat-kcurved;


  if(MixingPars.size()>0){
    //reference vector with positions of mixing gauge states
    vector<int> statesmixing;
    //Detect mixing gauge states for first mixing direction
    for(size_t j=0;j<Mbasis.size();++j){
      if(Mbasis[0][j]!=0)
	statesmixing.push_back(j);
    }
    //Detect mixing gauge states for other mixing direction (without repetition)
    for(size_t i=1;i<kmix ;++i){
      for(size_t j=0;j<Mbasis.size();++j)
	if(Mbasis[i][j]!=0 && j> statesmixing.back()){//avoid counting again states that are already on the list and insert the ones that were not present in the previous vector
	  statesmixing.push_back(j);	
	}
    }
    if(statesmixing.size()!=kmix){
      cerr << "ERROR in bool Potential::FindMixing(): Number of scalar states mixing ("<<statesmixing.size()<<") is different from the number of detected mixing directions ("<<kmix<<"), so user defined mixing matrix is ambigous!" << endl; 
    exit(-1);
    }
    
    //Generate parameters
    for(size_t i=0; i!=MixingPars.size();++i){
      MixingPars[i]=r(MixingPars.vgetmin(i),MixingPars.vgetmax(i));//Generate parameters uniformly in range decided in mathematica file
    }
    //Calculate mixing matrix according to user defined expressions
    MyInternalMixing(PhiPar,Phi,MixingPars,Mrot,r);

    // Rotate Mrot backwards on the right, to cancel out the arbitrary basis that was used in the eigen-directions search 
    for(size_t i=0;i<kmix;++i)
      for(size_t j=0;j<kmix;++j){
	Mrot2[i][j]=0;
	for(size_t p=0;p<kmix;++p)
	  Mrot2[i][j]+=Mrot[i][p]*Mbasis[j][statesmixing[p]];//Fill up temporary matrix with "right"-rotated matrix
      }
    for(size_t i=0;i<kmix;++i)
      for(size_t j=0;j<kmix;++j){
	Mrot[i][j]=Mrot2[i][j];//Set elements of the new rotation matrix
      }
  }
  else{
    //Initialize Mrot to zero
    for(size_t i=0; i!=kmix;++i)
      for(size_t j=0; j!=kmix;++j)
	Mrot[i][j]=0;
    
    int sumdims=0;
    for(size_t b=0;b!=BlockDims.size();++b){
      GenMrot(TempM,BlockDims[b],r);
      // for(size_t i=0;i<BlockDims[b];++i){
      //for(size_t j=0;j<BlockDims[b];++j)
      //  clog<< TempM[BlockDims[b]*i+j]<< "\t";
      //clog<< endl;
      //}
      //clog<<"----"<<endl;
      for(size_t i=0; i!=BlockDims[b];++i)
	for(size_t j=0; j!=BlockDims[b];++j)
	  Mrot[sumdims+i][sumdims+j]=TempM[BlockDims[b]*i+j];
      sumdims+=BlockDims[b];
    }

    //for(size_t i=0; i!=kmix;++i){
      //for(size_t j=0; j!=kmix;++j)
      //clog<< Mrot[i][j]<< "\t";
      //clog<< endl;
    //}
    //clog<<"----"<<endl;

    
    //GenerateMrot(d2Vnew,Mrot,orderdV,kmix,NIC,r);
    ConvertWithMrot(d2Vnew,Mrot,orderdV,kmix,NIC,r);
    //exit(-1);
  }

#ifdef DEBUG
    ////////////////////////CHECK!!!!!!!!!!!!!!!!!!!!
  clog << endl<< "****** Internal mixing matrix ****** " << endl;
  for(size_t i=0;i<kmix;++i){
    for(size_t j=0;j<kmix;++j)
      clog<<Mrot[i][j]<< "\t";
    clog<< endl;
  }
  clog<< endl;
#endif

  //unsigned int kcurved=d2Vnew.size()-(kflat-kminus)-kminus2;
  unsigned int Nij=(kmix*(kmix+1))/2+kcurved; //Number of equations in the system (lines)
  unsigned int NL=d2Vnew[0][0].size()-NIC+kmix+kcurved; //Number of variables (columns)
  


  //Re-sizing matriz that will be used for the calculations to the correct dimensions
  d2Vreduced.resize(Nij); 
  for(size_t i=0;i!=Nij;++i)
    d2Vreduced[i].resize(NL);

  ///Set first conditions corresponding to equaling to the mass squared (diagonal conditions)
  for(size_t i=0;i!=kmix+kcurved;++i){
    //Set columns corresponding to couplings to their values
    for(size_t A=0; A!=d2Vnew[0][0].size()-NIC; ++A)
      d2Vreduced[i][A]=d2Vnew[i][i][orderdV[A+NIC]];
    //Set remaining columns to a "diagonal" form where the masses squared appear qith coefficient -1
    for(size_t A=d2Vnew[0][0].size()-NIC; A!=NL; ++A){
      if(A-(d2Vnew[0][0].size()-NIC)==i)
	d2Vreduced[i][A]=-1e0;
      else
	d2Vreduced[i][A]=0e0;	
    }
  }

  ///Set other (off-diagonal) conditions from the mixed block
  unsigned int k=kmix+kcurved;
  for(size_t i=0;i!=kmix;++i){
    for(size_t j=i+1;j!=kmix; ++j){
      for(size_t A=0; A!=d2Vnew[0][0].size()-NIC; ++A)
	d2Vreduced[k][A]=d2Vnew[i][j][orderdV[A+NIC]];
      
      for(size_t A=d2Vnew[0][0].size()-NIC; A!=NL; ++A)
	d2Vreduced[k][A]=0e0;
      
      k++;
    }
  }

#ifdef DEBUG
  clog << endl<< "****** d2Vreduced final homogeneous system  ******" << endl;
  PrintM1(d2Vreduced);
#endif
  
  //Perform Gaussian elimination
  ////Initialise counter of horizontal continuations in gaussian elimination
  int hcont=0;
  ////Call Gaussian elimination routine to act on d2Vreduced
  hcont=GaussUpTriang(d2Vreduced);
  
#ifdef DEBUG
  /////Check: Print out matrix
  clog << endl << "****** Print out d2Vreduced after step 1 of gaussian elimination ********";
  PrintM1(d2Vreduced);
#endif

 /////////////////////////////////////
  // Case when the solution of the system is the null vector (massless theory with kinetic terms only --> discard)
  ///!!!!!!!!!!!!!!!!!!!!!introduced extra condition recently!!!! check!!!!
  if(hcont==0 && (Nij>=NL)){
    return false;
  }
  else{
    // Find position of variables to be eliminated in the second step, perform second step of Gaussian elimination

    GaussSolveUpTriang(d2Vreduced,NIC2,orderd2V,hcont);

#ifdef DEBUG
    /////////////////////////////////////////////////////////
    /////Check: Print out matrix
    clog << endl << "****** Print out d2Vreduced after step 2 of gaussian elimination ********";
    PrintM1(d2Vreduced);
    clog << "Number of independent linear conditions = " << NIC2 << endl;

#endif
  }

#ifdef DEBUG
  ///////////////////////////////////////////////
  ///////Check: Print out ordering of couplings
  clog << endl<< "******** Ordering of couplings ***********"<< endl;
  for(size_t A=0;A!=NL;++A){
    clog << "A = " << A << " d2Vreduced[" << A<<"] = " <<orderd2V[A] << endl;
  }
#endif
  ////////////////////////////////////////////////////////////

  //Fill integers with info about mixing and Goldstones
  Nmixed=kmix;
  NnoMixCurv=kcurved;
  NGold=d2Vnew.size()-NnoMixCurv-Nmixed;
  return true;
}

void Potential::a0treeIni(){

  vector< vector<int> > PowStates;
  PowStates.resize((Phi.size()*(Phi.size()+1))/2);
  for(size_t i=0; i!=(Phi.size()*(Phi.size()+1))/2;++i){
    PowStates[i].resize(Phi.size());
    for(size_t j=0; j!=Phi.size();++j){
      PowStates[i][j]=0;
      if(i<Phi.size() && i==j)
	PowStates[i][j]=2;
    }
  }

  size_t count=Phi.size();
  for(size_t i=0;i!=Phi.size();++i){
    for(size_t j=i+1;j<Phi.size();++j){
      PowStates[count][i]=1;
      PowStates[count][j]=1;
      ++count;
    }
  }

  // Now increment by the correct amount using all possible vertices

  for(size_t a=0; a!=Cp.size(); ++a){
    for(size_t nop=0;nop!=Mon[a].size();++nop){
      for(size_t i=0; i!=(Phi.size()*(Phi.size()+1))/2;++i){
	for(size_t j=0;j!=(Phi.size()*(Phi.size()+1))/2;++j){
	  double Prefac=1;
	  if(i<Phi.size())
	    Prefac/=sqrt(2);
	  if(j<Phi.size())
	    Prefac/=sqrt(2);	
	  //test if this vertex contributes by checking powers
	  for(size_t k=0;k!=Phi.size();++k){
	    int p=PowStates[i][k]+PowStates[j][k];
	    /////////////
	    if(Mon[a][nop][k]==p)
	      Prefac*=factorial(p);
	    else{
	      Prefac=0;
	      break;
	    }
	  }
	  a0tree[i][j][a]+=Prefac*Mon[a][nop].coeff_get()/16e0/acos(-1);
	}
      }
    }
  }

#ifdef DEBUG
  /////Check: Print out matrix
  std::clog << std::endl << "****** Printing out Matrices used later for tree level unitarity check ********";
  printa0tree();
#endif

}

bool Potential::CheckTreeUni(){
  //Compute the matrix of amplitudes for the current parameter space point
  size_t dim=(Phi.size()*(Phi.size()+1))/2;
  for(size_t i=0; i<dim;++i){
    for(size_t j=0;j<dim;++j){
      a0treeSum[dim*i+j]=0;
      for(size_t a=0; a!=Cp.size(); ++a){	
	a0treeSum[dim*i+j]+=a0tree[i][j][a]*Cp.L[a];
      }
    }
  }

  //Compute Eigenvalues (matrix is destroyed!)
  EigenSystemRsymNoSort(dim,a0treeSum,a0treeSum);

  for(size_t i=0;i<dim;++i){
    if(abs(a0treeSum[i])>0.5)
      return false;
  }
  return true;
}

void Potential::BoundTreeUni(){
  //Compute the matrix of amplitudes for the current parameter space point
  size_t dim=(Phi.size()*(Phi.size()+1))/2;
  for(size_t i=0; i<dim;++i){
    for(size_t j=0;j<dim;++j){
      for(size_t a=0; a!=Cp.size(); ++a){
	if(abs(a0tree[i][j][a]*Cp.L.vmin(a))>0.5){
#ifdef VERBOSE
	  clog<< "Reseting Lmin[" << a <<"] = "<< Cp.L.vmin(a) <<" based on tree level unitarity to Lmin[" << a <<"] = "; 
#endif
	  Cp.L.vmin(a)/=abs(2*Cp.L.vmin(a)*a0tree[i][j][a]);
#ifdef VERBOSE
	  clog<< Cp.L.vmin(a)<< endl;
#endif	  
	}
	if(abs(a0tree[i][j][a]*Cp.L.vmax(a))>0.5){

#ifdef VERBOSE
	  clog<< "Reseting Lmax[" << a <<"] = "<< Cp.L.vmax(a) <<" based on tree level unitarity to Lmax[" << a <<"] = ";
#endif
	  Cp.L.vmax(a)/=abs(2*Cp.L.vmax(a)*a0tree[i][j][a]);
#ifdef VERBOSE
	  clog<< Cp.L.vmax(a)<< endl;
#endif
	}
      }
    }
  }
}

bool Potential::CheckBounds(){
  for(size_t a=0;a!=Cp.size();++a)
    if(Cp.L[a]<Cp.L.vgetmin(a) || Cp.L[a]>Cp.L.vgetmax(a))
      return false;
  for(size_t i=0;i!=Cp.Mass.size();++i)
    if(Cp.Mass[i]<Cp.Mass.vgetmin(i) || Cp.Mass[i]>Cp.Mass.vgetmax(i))
      return false;
 
  return true;
}


void Potential::PrintHeader(){
  clog << endl<<"***********************************************" << endl<<"*********** ScannerS version 1.0.0 ************" << endl<< "***********************************************" << endl <<endl<<"---------- SUMMARY INFO FOR THIS RUN ---------- " <<endl<<endl<< "The potential you have entered is:"<< endl<< endl<< (*this)<< endl<<endl<< "The ranges of parameters to be scanned are:" << endl<< endl<< "Ranges for vacuum expectation values:"<< endl;
  for(size_t i=0;i!=Phi.size();++i)
    clog<< "phi"<< i<< " in ["<<Phi.vgetmin(i)<<":"<<Phi.vgetmax(i)<<"]"<< endl;

  clog <<endl<<"Ranges for couplings:"<< endl;
  for(size_t i=0;i!=Cp.L.size();++i)
    clog<< "L"<< i<< " in ["<<Cp.L.vgetmin(i)<<":"<<Cp.L.vgetmax(i)<<"]"<< endl;
  clog <<endl<<"Ranges for physical state masses:"<< endl;
  for(size_t i=0;i!=Cp.Mass.size();++i)
    clog<< "m"<< i<< " in ["<<Cp.Mass.vgetmin(i)<<":"<<Cp.Mass.vgetmax(i)<<"]"<< endl;

  clog<< endl;
  clog<< "The seed you are using for the random number generator is: " << (*this).r.seed<< endl<< endl;
}


//////////////////////////////////////////////////////////////////////////
////////////////// VERBOSE MODE ROUTINES /////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void Potential::printdV(){
  clog<< endl;
  for(std::vector< std::vector<double> >::iterator iter=dV.begin();iter!=dV.end();++iter){
    clog<< "| ";
    for(std::vector<double>::iterator ater=(*iter).begin();ater!=(*iter).end();++ater){
      clog<<scientific<<*ater<<"\t";
    }
    clog<<" |"<<endl;
  }
  clog<< endl;
}

void Potential::printd2V(){
  for(size_type a=0; a!=d2V[0][0].size();++a){
    clog<< " a = " <<a<<endl;
    for(size_type i=0;i!=d2V.size();++i){
      clog<< "| ";
      for(size_type j=0;j!=d2V.size();++j){
	clog<<scientific<<d2V[i][j][a]<<"\t";
      }
      clog<<" |"<<endl;
    }
    clog<< endl;
  }
  clog<< endl;
}

void Potential::printa0tree(){
  for(size_type a=0; a!=a0tree[0][0].size();++a){
    clog<< " a = " <<a<<endl;
    for(size_type i=0;i!=a0tree.size();++i){
      clog<< "| ";
      for(size_type j=0;j!=a0tree.size();++j){
	clog<<scientific<<a0tree[i][j][a]<<"\t";
      }
      clog<<" |"<<endl;
    }
    clog<< endl;
  }
  clog<< endl;
}



